import gzip
import subprocess
import struct
import bpy
import copy
from math import sqrt
from collections import defaultdict
from collections import namedtuple

print("---start")

Vector3D = namedtuple('Vector3D','x y z')
Triangle = namedtuple('Triangle','v0 v1 v2')
TriangleTex = namedtuple('TriangleTex','v0 v1 v2 tex')
UVcoords = namedtuple('UVcoords','u v')

CombiVertex = namedtuple('CombiVertex','position normal uvs context')
Vertex = namedtuple('Vertex','position normal uvs')
ContextStrips = namedtuple('ContextStrips','context strips')

def stripify(faces, params = "", stripypath = "C:\\cygwin\\tristripers\\Stripy\\tristrip.exe"):
    # Start the stripifier process
    assert len(faces) > 0
    
    if params == "":
        p = subprocess.Popen(stripypath, stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=False)
    else:
        cmdline = [stripypath]
        cmdline.extend(params.split())
        p = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=False)
        #p = subprocess.Popen([stripypath, params], stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=False)
    
    # Send over each face
    for face in faces:
        #print("{0[0]} {0[1]} {0[2]} ".format(face))
        p.stdin.write(bytes("{0[0]} {0[1]} {0[2]} ".format(face),'UTF-8'))
    
    # Send terminator
    p.stdin.write(bytes("-1\n",'UTF-8'))

    # Get results
    lines = p.communicate()[0].decode("utf-8").splitlines()

    if 1 == 0:
        # dump results for debugging
        file = open("C:\\cygwin\\tris.txt", "wb")
        for line in lines:
            file.write(bytes(line + '\n','UTF-8'))
        file.close()


    # Convert to lists of ints
    strips = []
    fans = 0
    for sidx in range(1, int(lines[0])*3+1, 3):
        striptype = int(lines[sidx])
        if striptype != 1:
            fans = fans + 1
        strips.append([])
        strips[-1].extend(map(int, lines[sidx+2].split()))

    print("Fans: ",fans)
    assert fans == 0
    return strips


###########################################################################
###########################################################################
#                   PACKING
###########################################################################
###########################################################################

def pack_header(strip_cnt,strip_len_start,vertex_start):
    return struct.pack('<iii20x',strip_cnt,strip_len_start,vertex_start)
def pack_strip_len(strip_len):
    return struct.pack('<i',strip_len)
def pack_vert(x,y,z,nx=0.0,ny=0.0,nz=1.0,u=0.0,v=0.0):
    return struct.pack('<8f',x,y,z,u,nx,ny,nz,-v)
def pack_nvert(x,y,z,nx=0.0,ny=0.0,nz=1.0, tx=0.0,ty=0.0,tz=0.0, btx=0.0,bty=0.0,btz=0.0,  u=0.0,v=0.0):
    #return struct.pack('<8f',x,-y,-z,u,nx,-ny,-nz,-v)
    return struct.pack('<8f3f4x3f4x',x,y,z,u,nx,ny,nz,-v, tx,ty,tz, btx,bty,btz)

def pack_vert2(twovects):
    return pack_vert(twovects[0][0],twovects[0][1],twovects[0][2],twovects[1][0],twovects[1][1],twovects[1][2])
def pack_vert3(v):
    return pack_vert(v.position[0],v.position[1],v.position[2],v.normal[0],v.normal[1],v.normal[2],v.uvs[0],v.uvs[1])
def pack_norm(v):
    return 
#this is what pack_vert3 does:
#def pack_vert4(threevects):
#    return pack_vert(*threevects.normal,*threevects.normal,*threevects.uv)



###########################################################################
###########################################################################
#                   BLENDER DATA FETCHING
###########################################################################
###########################################################################

def tessed_copy(mesh):
#    m = mesh.to_mesh(bpy.context.scene, True, 'PREVIEW')
    m = mesh.data
    m.update(True,True)
    return m

def gettriangles(model):
    polys = []
    for i,f in enumerate(model.tessfaces):
        tex = 0
        texstr = "gouraud-intensity"
        if len(model.tessface_uv_textures):
            if i < len(model.tessface_uv_textures[0].data):
                if model.tessface_uv_textures[0].data[i].image:
                    texstr = model.tessface_uv_textures[0].data[i].image.name
        
        #I don't think tessfaces should have polygons with more than three sides, but...
        if len(f.vertices) == 3:
            polys.append( (TriangleTex(f.vertices[0], f.vertices[1], f.vertices[2], texstr)) )
        elif len(f.vertices) == 4:
            polys.append( (TriangleTex(f.vertices[0], f.vertices[1], f.vertices[2],texstr)) )
            polys.append( (TriangleTex(f.vertices[0], f.vertices[2], f.vertices[3],texstr)) )
                
        else:
            print("HEY")
            print(len(f))
            assert 0
    return polys


def getpositions(model):
    # create list containing vertex positions from a tessfaced model
    pos = []
    for p in model.vertices:
        pos.append( Vector3D(*p.co) )
    return pos

def getnormals(model):
    # create list containing vertex normals from a tessfaced model
    norms = []
    for p in model.vertices:
        norms.append( Vector3D(*p.normal) )
    return norms

def getuvs(model):
    # create a list containing uvs. if no textures, return empty list
    uvs = []
    if not model.tessface_uv_textures:
        return uvs
    
    for f in model.tessface_uv_textures[0].data:
        if len(f.uv) == 3:
            uvs.append( ( UVcoords(*f.uv[0]),  UVcoords(*f.uv[1]),  UVcoords(*f.uv[2]) ) )
        else:
            uvs.append( ( UVcoords(*f.uv[0]),  UVcoords(*f.uv[1]),  UVcoords(*f.uv[2]) ) )
            uvs.append( ( UVcoords(*f.uv[0]),  UVcoords(*f.uv[2]),  UVcoords(*f.uv[3]) ) )
    return uvs

###########################################################################
###########################################################################
#                   BLENDER DATA PROCESSING
###########################################################################
###########################################################################

def make_uv_triangles(model, tris, poss, norms, uvs):
    ''' output format:
        ntris is an array containing triangles. each triangle is made of three vertex indices.        
        nverts is an array containing vertices. each vertex is made of a position, normal, uv, and optional texture
    '''
    ntris = []
    nverts = []
    vdict = dict()
    
    def combine(pos,norm,uv,tex):
        if tex == 0 or tex == None:
            return ( CombiVertex(pos[0],norm[0],uv[0],"gouraud-intensity"), CombiVertex(pos[1],norm[1],uv[1],"gouraud-intensity"), CombiVertex(pos[2],norm[2],uv[2],"gouraud-intensity") )
        else:
            return ( CombiVertex(pos[0],norm[0],uv[0],tex), CombiVertex(pos[1],norm[1],uv[1],tex),
                CombiVertex(pos[2],norm[2],uv[2],tex) )
    
    def add_vert(v):
        if v not in vdict:
            vdict[v] = len(nverts)
            nverts.append(v)
            return len(vdict)-1
        else:
            return vdict[v]
    
    for triidx,tri in enumerate(tris):
        if uvs:
            tuvs = uvs[triidx]
            if len(model.tessface_uv_textures[0].data) > triidx:
                tex = model.tessface_uv_textures[0].data[triidx].image  #TODO test and make sure this works
            else:
                tex = 0
        else:
            tuvs = ( UVcoords(0,0), UVcoords(0,0), UVcoords(0,0) )
            tex = 0
        tposs = ( poss[tri[0]], poss[tri[1]], poss[tri[2]] )
        tnorms = ( norms[tri[0]], norms[tri[1]], norms[tri[2]] )
        
        t = combine(tposs,tnorms,tuvs,tri.tex)

        ntris.append( (add_vert(t[0]), add_vert(t[1]), add_vert(t[2])) )
    
    return ntris, nverts

def makestrip_tex(strips, verts):
    #take generated strip indices and combine with full vertex data
    compstrips = []
    for strip in strips:
        compstrips.append([])
        for v in strip:
            compstrips[-1].append(verts[v])
    return compstrips

###########################################################################
###########################################################################
#                   WRITING
###########################################################################
###########################################################################
def write_model_ex(name,processed):
    assert len(processed) > 0
    
    strip_cnt = len(processed)
    strip_len_sec = bytes()
    vertex_sec = bytes()
    
    for strip in processed:
        strip_len_sec = strip_len_sec + pack_strip_len(len(strip))
        for vertex in strip:
            vertex_sec = vertex_sec + pack_vert3(vertex)
    
    #pad out
    def pad_up(bytestr,place):
        if len(bytestr) % place:
            return bytestr + bytes(place - (len(bytestr) % place))
        return bytestr
    
    strip_len_sec = pad_up(strip_len_sec,32)
    #write it out
    print("Writing...")
    file = gzip.open(name, "wb")
    file.write(pack_header(strip_cnt,32,32+len(strip_len_sec)))
    file.write(strip_len_sec)
    file.write(vertex_sec)
    file.close()


foo = tessed_copy(bpy.data.objects["Deadpool"])

footri = gettriangles(foo)
foouvs = getuvs(foo)
foopos = getpositions(foo)
foonorms = getnormals(foo)

print("Model: ", foo)
print("T: ", len(footri), " U: ", len(foouvs), " P: ", len(foopos), " N: ", len(foonorms))

print("Building triangles...")
nt, nv = make_uv_triangles(foo, footri,foopos,foonorms,foouvs)
print("Number of triangles: ", len(nt))
print("Number of unique vertices: ", len(nv))

print("Generating strips...")
strips = stripify(nt,"-cs 100")
print("Number of strips: ", len(strips))
t = 0
for s in strips:
    t = t + len(s)
print("Total vertices: ", t)
print("Inefficiency: ", t/len(nt))
processed = makestrip_tex(strips,nv)

write_model_ex("C:\\cygwin\\deadpool.bin.gz",processed)

print("ok")