#include "profile.h"
#include "debug.h"
#include "watchdog.h"

#define CLOCKS_PER_SECOND	(200000000.0f)

typedef struct {
	unsigned char addr[3];
} small_address_t;

static kthread_t *profile_thread = NULL;
static volatile int profile_inited = 0;
static volatile int profiling = 0;

static small_address_t *profile_store = NULL;
static int profile_pos = 0;
static int profile_max = 0;

float PfSecs(profile_t time)
{
	return time / 1000000.0f;
}

float PfCycles(profile_t time)
{
	return CLOCKS_PER_SECOND * PfSecs(time);
}

float PfCycLoop(profile_t time, int loops)
{
	return PfCycles(time) / loops;
}

void ProfileColorPoint(int red, int green, int blue)
{
#ifdef _arch_dreamcast
	vid_border_color(red, green, blue);
#endif
}

void profile_init(size_t size)
{
	assert(watchdog_running());
	assert(!profile_inited);
	
	profile_pos = 0;
	profiling = 0;
	profile_thread = thd_current;
	profile_max = size;
	if (profile_store == NULL)
		profile_store = malloc(sizeof(small_address_t) * profile_max);
	
	profile_inited = 1;
}

void profile_reset(void)
{
	assert(profile_inited);
	profile_pos = 0;
	profiling = 0;
}

void profile_shutdown(void)
{
	assert(profile_inited);
	
	profile_thread = NULL;
	profile_inited = 0;
	profiling = 0;
	if (profile_store)
		free(profile_store);
	profile_store = NULL;
	profile_pos = 0;
	profile_max = 0;
}

void profile_go(void)
{
	assert(profile_inited);
	assert(!profiling);
	profiling = 1;
}

void profile_stop(void)
{
	assert(profile_inited);
	profiling = 0;
}

int profile_running(void)
{
	return profiling;
}

void profile_record_pc(void)
{
	if (profile_pos >= profile_max) {
		profiling = 0;
		DBOS("Profile records filled!\n");
	} else {
		unsigned int pc = (unsigned int)profile_thread->context.pc;
		
		profile_store[profile_pos].addr[0] = pc & 0xff;
		profile_store[profile_pos].addr[1] = (pc>>8) & 0xff;
		profile_store[profile_pos].addr[2] = (pc>>16) & 0xff;
		profile_pos++;
	}
}

int profile_save(const char *fname, int append)
{
	assert(profile_inited);
	assert(!profiling);
	
	DBOS("Writing %i profile samples to %s... ", profile_pos,fname);
	
	FILE *of;
	int result = 1;
	size_t written = 0;
	
	of = fopen(fname, append ? "a" : "w");
	if (of != NULL) {
		//written = fwrite(&profile_store[0], sizeof(small_address_t), profile_pos, of);
		int i;
		for(i = 0; i < profile_pos; i++) {
			uint32 addr = (profile_store[i].addr[2] << 16) | (profile_store[i].addr[1] << 8) | profile_store[i].addr[0];
			written += fprintf(of,"8C%06X\n",addr);
		}
		//written = fwrite(&profile_store[0], sizeof(small_address_t), profile_pos, of);
		fclose(of);
		//if (written == profile_pos)
			result = 0;
	}
	
	DBOS("Done! %i bytes (%i, %i)\n",written,result,ferror(of));
	
	return result;
}

